from copy import deepcopy
import six

from .base import (
    PlanOutOp,
    PlanOutOpSimple,
    PlanOutOpBinary,
    PlanOutOpUnary,
    PlanOutOpCommutative,
)

import planout.ops.utils as ops


def indent(s, n=1):
    l = [("  " * n) + i for i in s.split('\n')]
    return '\n'.join(l)


class Literal(PlanOutOp):

    def execute(self, mapper):
        return self.getArgMixed('value')

    def pretty(self):
        return self.getArgMixed('value')


class Get(PlanOutOp):

    def execute(self, mapper):
        return mapper.get(self.getArgString('var'))

    def pretty(self):
        return self.getArgString('var')


class Seq(PlanOutOp):

    def execute(self, mapper):
        for op in self.getArgList('seq'):
            mapper.evaluate(op)

    def pretty(self):
        l = [ops.Operators.pretty(v) for v in self.getArgList('seq')]
        return '\n'.join(l)


class Set(PlanOutOp):

    def execute(self, mapper):
        var, value = self.getArgString('var'), self.getArgMixed('value')
        if mapper.has_override(var):
            return

        # if the value is operator, add the name of the variable as a salt if no
        # salt is provided.
        if ops.Operators.isOperator(value) and 'salt' not in value:
            value['salt'] = var

        # if we are setting the special variable, experiment_salt, update mapper
        # object accordingly with the new experiment-level salt
        if var == 'experiment_salt':
            mapper.experiment_salt = value

        mapper.set(var, mapper.evaluate(value))

    def pretty(self):
        strp = ops.Operators.pretty(self.getArgMixed('value'))
        return "%s = %s;" % (self.getArgString('var'), strp)


class Return(PlanOutOp):

    def execute(self, mapper):
        # if script calls return; or return();, assume the unit is in the
        # experiment
        value = mapper.evaluate(self.getArgMixed('value'))
        in_experiment = True if value else False
        raise ops.StopPlanOutException(in_experiment)


class Array(PlanOutOp):

    def execute(self, mapper):
        return [mapper.evaluate(value) for value in self.getArgList('values')]

    def pretty(self):
        l = [ops.Operators.pretty(v) for v in self.getArgList('values')]
        f = "[%s]" % ', '.join(l)
        return f


class Map(PlanOutOpSimple):

    def simpleExecute(self):
        m = deepcopy(self.args)
        del m['op']
        del m['salt']
        return m


class Coalesce(PlanOutOp):

    def execute(self, mapper):
        for x in self.getArgList('values'):
            eval_x = mapper.evaluate(x)
            if eval_x is not None:
                return eval_x
        return None

    def pretty(self):
        values = Operators.strip_array(self.getArgList('values'))
        pretty_c = [Operators.pretty(i) for i in values]
        return 'coalesce(%s)' % ', '.join(pretty_c)


class Index(PlanOutOpSimple):

    def simpleExecute(self):
        # returns value at index if it exists, returns None otherwise.
        # works with both lists and dictionaries.
        base, index = self.getArgIndexish('base'), self.getArgMixed('index')
        if type(base) is list:
            if index >= 0 and index < len(base):
                return base[index]
            else:
                return None
        else:
            # assume we have a dictionary
            return base.get(index)

        return self.getArgIndexish('base')[self.getArgMixed('index')]

    def pretty(self):
        b = ops.Operators.pretty(self.getArgIndexish('base'))
        i = ops.Operators.pretty(self.getArgMixed('index'))
        return "%s[%s]" % (b, i)


class Cond(PlanOutOp):

    def execute(self, mapper):
        for i in self.getArgList('cond'):
            if_clause, then_clause = i['if'], i['then']
            if mapper.evaluate(if_clause):
                return mapper.evaluate(then_clause)

    def pretty(self):
        pretty_str = ""
        first_if = True
        for i in self.getArgList('cond'):
            if_clause, then_clause = i['if'], i['then']
            if if_clause == 'true':
                pretty_str += 'else {\n'
            else:
                prefix = 'if(%s) {\n' if first_if else 'else if(%s) {\n'
                pretty_str += prefix % ops.Operators.pretty(if_clause)
            pretty_str += indent(ops.Operators.pretty(then_clause)) + '\n}'
        return pretty_str


class And(PlanOutOp):

    def execute(self, mapper):
        for clause in self.getArgList('values'):
            if not mapper.evaluate(clause):
                return False
        return True

    def pretty(self):
        pretty_c = [Operators.pretty(i) for i in self.getArgList('values')]
        return '&& '.join(pretty_c)


class Or(PlanOutOp):

    def execute(self, mapper):
        for clause in self.getArgList('values'):
            if mapper.evaluate(clause):
                return True
        return False

    def pretty(self):
        pretty_c = [Operators.pretty(c) for c in self.getArgList('values')]
        return '|| '.join(pretty_c)


class Product(PlanOutOpCommutative):

    def commutativeExecute(self, values):
        return six.moves.reduce(lambda x, y: x * y, values)

    def pretty(self):
        values = Operators.strip_array(self.getArgList('values'))
        pretty_c = [Operators.pretty(i) for i in values]
        return ' * '.join(pretty_c)


class Sum(PlanOutOpCommutative):

    def commutativeExecute(self, values):
        return sum(values)

    def pretty(self):
        pretty_c = [Operators.pretty(c) for c in self.getArgList('values')]
        return '+ '.join(pretty_c)


class Equals(PlanOutOpBinary):

    def getInfixString(self):
        return "=="

    def binaryExecute(self, left, right):
        return left == right


class GreaterThan(PlanOutOpBinary):

    def binaryExecute(self, left, right):
        return left > right


class LessThan(PlanOutOpBinary):

    def binaryExecute(self, left, right):
        return left < right


class LessThanOrEqualTo(PlanOutOpBinary):

    def binaryExecute(self, left, right):
        return left <= right


class GreaterThanOrEqualTo(PlanOutOpBinary):

    def binaryExecute(self, left, right):
        return left >= right


class Mod(PlanOutOpBinary):

    def binaryExecute(self, left, right):
        return left % right


class Divide(PlanOutOpBinary):

    def binaryExecute(self, left, right):
        return float(left) / float(right)


class Round(PlanOutOpUnary):

    def unaryExecute(self, value):
        return round(value)


class Not(PlanOutOpUnary):

    def unaryExecute(self, value):
        return not value

    def getUnaryString():
        return '!'


class Negative(PlanOutOpUnary):

    def unaryExecute(self, value):
        return 0 - value

    def getUnaryString():
        return '-'


class Min(PlanOutOpCommutative):

    def commutativeExecute(self, values):
        return min(values)


class Max(PlanOutOpCommutative):

    def commutativeExecute(self, values):
        return max(values)


class Length(PlanOutOpUnary):

    def unaryExecute(self, value):
        return len(value)



# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

from .ops.random import *
from .ops.base import PlanOutOp
from collections.abc import MutableMapping


# The Assignment class is the main work horse that lets you to execute
# random operators using the names of variables being assigned as salts.
# It is a MutableMapping, which means it plays nice with things like Flask
# template renders.
class Assignment(MutableMapping):

    """
    A mutable mapping that contains the result of an assign call.
    """

    def __init__(self, experiment_salt, overrides={}):
        self.experiment_salt = experiment_salt
        self._overrides = overrides.copy()
        self._data = overrides.copy()
        self.salt_sep = '.' # separates unit from experiment/variable salt

    def evaluate(self, value):
        return value

    def get_overrides(self):
        return self._overrides

    def set_overrides(self, overrides):
        # maybe this should be a deep copy?
        self._overrides = overrides.copy()
        for var in self._overrides:
            self._data[var] = self._overrides[var]

    def __setitem__(self, name, value):
        if name in ('_data', '_overrides', 'salt_sep', 'experiment_salt'):
            self.__dict__[name] = value
            return

        if name in self._overrides:
            return

        if isinstance(value, PlanOutOpRandom):
            if 'salt' not in value.args:
                value.args['salt'] = name
            self._data[name] = value.execute(self)
        else:
            self._data[name] = value

    __setattr__ = __setitem__

    def __getitem__(self, name):
        if name in ('_data', '_overrides', 'experiment_salt'):
            return self.__dict__[name]
        else:
            return self._data[name]

    __getattr__ = __getitem__

    def __delitem__(self, name):
        del self._data[name]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)



import json
import six

class StopPlanOutException(Exception):

    """Exception that gets raised when "return" op is evaluated"""

    def __init__(self, in_experiment):
        self.in_experiment = in_experiment


class Operators():
    """Singleton class for inspecting and registering operators"""

    @staticmethod
    def initFactory():
        from . import core, random
        Operators.operators = {
            "literal": core.Literal,
            "get": core.Get,
            "seq": core.Seq,
            "set": core.Set,
            "return": core.Return,
            "index": core.Index,
            "array": core.Array,
            "map": core.Map,
            "equals": core.Equals,
            "cond": core.Cond,
            "and": core.And,
            "or": core.Or,
            ">": core.GreaterThan,
            "<": core.LessThan,
            ">=": core.GreaterThanOrEqualTo,
            "<=": core.LessThanOrEqualTo,
            "%": core.Mod,
            "/": core.Divide,
            "not": core.Not,
            "round": core.Round,
            "negative": core.Negative,
            "min": core.Min,
            "max": core.Max,
            "length": core.Length,
            "coalesce": core.Coalesce,
            "product": core.Product,
            "sum": core.Sum,
            "randomFloat": random.RandomFloat,
            "randomInteger": random.RandomInteger,
            "bernoulliTrial": random.BernoulliTrial,
            "bernoulliFilter": random.BernoulliFilter,
            "uniformChoice": random.UniformChoice,
            "weightedChoice": random.WeightedChoice,
            "sample": random.Sample,
            "fastSample": random.FastSample
        }

    @staticmethod
    def registerOperators(operators):
        for op, obj in six.iteritems(operators):
            assert op not in Operators.operators
            Operators.operators[op] = operators[op]

    @staticmethod
    def isOperator(op):
        return type(op) is dict and "op" in op

    @staticmethod
    def operatorInstance(params):
        op = params['op']
        assert (op in Operators.operators), "Unknown operator: %s" % op
        return Operators.operators[op](**params)

    @staticmethod
    def prettyParamFormat(params):
        ps = [p + '=' + Operators.pretty(params[p])
              for p in params if p != 'op']
        return ', '.join(ps)

    @staticmethod
    def strip_array(params):
        if type(params) is list:
            return params
        if type(params) is dict and params.get('op', None) == 'array':
            return params['values']
        else:
            return params

    @staticmethod
    def pretty(params):
        if Operators.isOperator(params):
            try:
                # if an op is invalid, we may not be able to pretty print it
                my_pretty = Operators.operatorInstance(params).pretty()
            except:
                my_pretty = params
            return my_pretty
        elif type(params) is list:
            return '[%s]' % ', '.join([Operators.pretty(p) for p in params])
        else:
            return json.dumps(params)



# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

from copy import deepcopy
from .ops.utils import Operators, StopPlanOutException
from .assignment import Assignment


Operators.initFactory()


class Interpreter(object):

    """PlanOut interpreter"""

    def __init__(self, serialization, experiment_salt='global_salt',
                 inputs={}, environment=None):
        self._serialization = serialization
        if environment is None:
            self._env = Assignment(experiment_salt)
        else:
            self._env = environment
        self.experiment_salt = self._experiment_salt = experiment_salt
        self._evaluated = False
        self._in_experiment = True
        self._inputs = inputs.copy()

    def register_operators(self, operators):
        Operators.registerOperators(operators)
        return self

    def get_params(self):
        """Get all assigned parameter values from an executed interpreter script"""
        # evaluate code if it hasn't already been evaluated
        if not self._evaluated:
            try:
                self.evaluate(self._serialization)
            except StopPlanOutException as e:
                # StopPlanOutException is raised when script calls "return", which
                # short circuits execution and sets in_experiment
                self._in_experiment = e.in_experiment
            self._evaluated = True
        return self._env

    @property
    def in_experiment(self):
        return self._in_experiment

    @property
    def salt_sep(self):
        return self._env.salt_sep

    def set_env(self, new_env):
        """Replace the current environment with a dictionary"""
        self._env = deepcopy(new_env)
        # note that overrides are inhereted from new_env
        return self

    def has(self, name):
        """Check if a variable exists in the PlanOut environment"""
        return name in self._env

    def get(self, name, default=None):
        """Get a variable from the PlanOut environment"""
        return self._env.get(name, self._inputs.get(name, default))

    def set(self, name, value):
        """Set a variable in the PlanOut environment"""
        self._env[name] = value
        return self

    def set_overrides(self, overrides):
        """
        Sets variables to maintain a frozen state during the interpreter's
        execution. This is useful for debugging PlanOut scripts.
        """
        self._env.set_overrides(overrides)
        return self

    def get_overrides(self):
        """Get a dictionary of all overrided values"""
        return self._env.get_overrides()

    def has_override(self, name):
        """Check to see if a variable has an override."""
        return name in self.get_overrides()

    def evaluate(self, planout_code):
        """Recursively evaluate PlanOut interpreter code"""
        # if the object is a PlanOut operator, execute it it.
        if type(planout_code) is dict and 'op' in planout_code:
            return Operators.operatorInstance(planout_code).execute(self)
        # if the object is a list, iterate over the list and evaluate each
        # element
        elif type(planout_code) is list:
            return [self.evaluate(i) for i in planout_code]
        else:
            return planout_code  # data is a literal



import hashlib
import six
from .base import PlanOutOpSimple


class PlanOutOpRandom(PlanOutOpSimple):
    LONG_SCALE = float(0xFFFFFFFFFFFFFFF)

    def getUnit(self, appended_unit=None):
        unit = self.getArgMixed('unit')
        if type(unit) is not list:
            unit = [unit]
        if appended_unit is not None:
            unit += [appended_unit]
        return unit

    def getHash(self, appended_unit=None):
        if 'full_salt' in self.args:
            full_salt = self.getArgString('full_salt') + '.'  # do typechecking
        else:
            full_salt = '%s.%s%s' % (
                self.mapper.experiment_salt,
                self.getArgString('salt'),
                self.mapper.salt_sep)

        unit_str = '.'.join(map(str, self.getUnit(appended_unit)))
        hash_str = '%s%s' % (full_salt, unit_str)
        if not isinstance(hash_str, six.binary_type):
            hash_str = hash_str.encode("ascii")
        return int(hashlib.sha1(hash_str).hexdigest()[:15], 16)

    def getUniform(self, min_val=0.0, max_val=1.0, appended_unit=None):
        zero_to_one = self.getHash(appended_unit) / PlanOutOpRandom.LONG_SCALE
        return min_val + (max_val - min_val) * zero_to_one


class RandomFloat(PlanOutOpRandom):

    def simpleExecute(self):
        min_val = self.getArgFloat('min')
        max_val = self.getArgFloat('max')

        return self.getUniform(min_val, max_val)


class RandomInteger(PlanOutOpRandom):

    def simpleExecute(self):
        min_val = self.getArgInt('min')
        max_val = self.getArgInt('max')

        return min_val + self.getHash() % (max_val - min_val + 1)


class BernoulliTrial(PlanOutOpRandom):

    def simpleExecute(self):
        p = self.getArgNumeric('p')
        assert p >= 0 and p <= 1.0, \
            '%s: p must be a number between 0.0 and 1.0, not %s!' \
            % (self.__class__, p)

        rand_val = self.getUniform(0.0, 1.0)
        return 1 if rand_val <= p else 0


class BernoulliFilter(PlanOutOpRandom):

    def simpleExecute(self):
        p = self.getArgNumeric('p')
        values = self.getArgList('choices')
        assert p >= 0 and p <= 1.0, \
            '%s: p must be a number between 0.0 and 1.0, not %s!' \
            % (self.__class__, p)

        if len(values) == 0:
            return []
        return [i for i in values if self.getUniform(0.0, 1.0, i) <= p]


class UniformChoice(PlanOutOpRandom):

    def simpleExecute(self):
        choices = self.getArgList('choices')

        if len(choices) == 0:
            return []
        rand_index = self.getHash() % len(choices)
        return choices[rand_index]


class WeightedChoice(PlanOutOpRandom):

    def simpleExecute(self):
        choices = self.getArgList('choices')
        weights = self.getArgList('weights')

        if len(choices) == 0:
            return []
        cum_weights = dict(enumerate(weights))
        cum_sum = 0.0
        for index in cum_weights:
            cum_sum += cum_weights[index]
            cum_weights[index] = cum_sum
        stop_value = self.getUniform(0.0, cum_sum)
        for index in cum_weights:
            if stop_value <= cum_weights[index]:
                return choices[index]


class BaseSample(PlanOutOpRandom):

    def copyChoices(self):
        return [x for x in self.getArgList('choices')]

    def getNumDraws(self, choices):
        if 'draws' in self.args:
            num_draws = self.getArgInt('draws')
            assert num_draws <= len(choices), \
                "%s: cannot make %s draws when only %s choices are available" \
                % (self.__class__, num_draws, len(choices))
            return num_draws
        else:
            return len(choices)

class FastSample(BaseSample):

    def simpleExecute(self):
        choices = self.copyChoices()
        num_draws = self.getNumDraws(choices)
        stopping_point = len(choices) - num_draws

        for i in six.moves.range(len(choices) - 1, 0, -1):
            j = self.getHash(i) % (i + 1)
            choices[i], choices[j] = choices[j], choices[i]
            if stopping_point == i:
                return choices[i:]
        return choices[:num_draws]

class Sample(BaseSample):

    def simpleExecute(self):
        choices = self.copyChoices()
        num_draws = self.getNumDraws(choices)

        for i in six.moves.range(len(choices) - 1, 0, -1):
            j = self.getHash(i) % (i + 1)
            choices[i], choices[j] = choices[j], choices[i]
        return choices[:num_draws]



# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

from collections import Counter
import unittest
from math import sqrt
import six

from planout.ops.random import *
from planout.assignment import Assignment

# decorator for quickly constructing PlanOutKit experiments


def experiment_decorator(name):
    def wrap(f):
        def wrapped_f(**kwargs):
            params = Assignment(name)
            return f(params, **kwargs)
        return wrapped_f
    return wrap


class TestRandomOperators(unittest.TestCase):
    # z_{\alpha/2} for \alpha=0.001, e.g., 99.9% CI: qnorm(1-(0.001/2))
    z = 3.29

    @staticmethod
    def valueMassToDensity(value_mass):
        """convert value_mass dictionary to a density"""
        values, ns = zip(*value_mass)
        ns_sum = float(sum(ns))
        value_density = dict(zip(values, [i / ns_sum for i in ns]))
        return value_density

    def distributionTester(self, func, value_mass, N=1000):
        """Make sure an experiment object generates the desired frequencies"""
        # run N trials of f() with input i
        xs = [func(i=i).get('x') for i in six.moves.range(N)]
        value_density = TestRandomOperators.valueMassToDensity(value_mass)

        # test outcome frequencies against expected density
        self.assertProbs(xs, value_density, float(N))

    def assertProbs(self, xs, value_density, N):
        """Assert a list of values has the same density as value_density"""
        hist = Counter(xs)

        # do binomial test of proportions for each item
        for i in hist:
            self.assertProp(hist[i] / N, value_density[i], N)

    def assertProp(self, observed_p, expected_p, N):
        """Does a test of proportions"""
        # normal approximation of binomial CI.
        # this should be OK for large N and values of p not too close to 0 or
        # 1.
        se = TestRandomOperators.z * sqrt(expected_p * (1 - expected_p) / N)
        self.assertTrue(abs(observed_p - expected_p) <= se)

    def test_salts(self):
        """Test salting behavior"""
        i = 20
        a = Assignment('assign_salt_a')

        # assigning variables with different names and the same unit should yield
        # different randomizations, when salts are not explicitly specified
        a.x = RandomInteger(min=0, max=100000, unit=i)
        a.y = RandomInteger(min=0, max=100000, unit=i)
        self.assertTrue(a.x != a.y)

        # when salts are specified, they act the same way auto-salting does
        a.z = RandomInteger(min=0, max=100000, unit=i, salt='x')
        self.assertTrue(a.x == a.z)

        # when the Assignment-level salt is different, variables with the same
        # name (or salt) should generally be assigned to different values
        b = Assignment('assign_salt_b')
        b.x = RandomInteger(min=0, max=100000, unit=i)
        self.assertTrue(a.x != b.x)

        # when a full salt is specified, only the full salt is used to do
        # hashing
        a.f = RandomInteger(min=0, max=100000, unit=i, full_salt='fs')
        b.f = RandomInteger(min=0, max=100000, unit=i, full_salt='fs')
        self.assertTrue(a.f == b.f)
        a.f = RandomInteger(min=0, max=100000, unit=i, full_salt='fs2')
        b.f = RandomInteger(min=0, max=100000, unit=i, full_salt='fs2')
        self.assertTrue(a.f == b.f)

    def test_bernoulli(self):
        """Test bernoulli trial"""

        # returns experiment function with x = BernoulliTrial(p) draw
        # experiment salt is p
        def bernoulliTrial(p):
            @experiment_decorator(p)
            def exp_func(e, i):
                e.x = BernoulliTrial(p=p, unit=i)
                return e
            return exp_func

        self.distributionTester(bernoulliTrial(0.0), ((0, 1), (1, 0)))
        self.distributionTester(bernoulliTrial(0.1), ((0, 0.9), (1, 0.1)))
        self.distributionTester(bernoulliTrial(1.0), ((0, 0), (1, 1)))

    def test_uniform_choice(self):
        """Test uniform choice"""

        # returns experiment function with x = UniformChoice(c) draw
        # experiment salt is a string version of c
        def uniformChoice(c):
            str_c = ','.join(map(str, c))

            @experiment_decorator(str_c)
            def exp_func(e, i):
                e.x = UniformChoice(choices=c, unit=i)
                return e
            return exp_func

        self.distributionTester(uniformChoice(['a']), [('a', 1)])
        self.distributionTester(
            uniformChoice(['a', 'b']), (('a', 1), ('b', 1)))
        self.distributionTester(
            uniformChoice([1, 2, 3, 4]), ((1, 1), (2, 1), (3, 1), (4, 1)))

    def test_weighted_choice(self):
        """Test weighted choice"""

        # returns experiment function with x = WeightedChoice(c,w) draw
        # experiment salt is a string version of weighted_dict's keys
        def weightedChoice(weight_pairs):
            c, w = zip(*weight_pairs)

            @experiment_decorator(','.join(map(str, w)))
            def exp_func(e, i):
                e.x = WeightedChoice(choices=c, weights=w, unit=i)
                return e
            return exp_func

        d = (('a', 1),)
        self.distributionTester(weightedChoice(d), d)
        d = (('a', 1), ('b', 2))
        self.distributionTester(weightedChoice(d), d)
        d = ((('a', 0), ('b', 2), ('c', 0)))
        self.distributionTester(weightedChoice(d), d)

        # we should be able to repeat the same choice multiple times
        # in weightedChoice(). in this case we repeat 'a'.
        da = ((('a', 1), ('b', 2), ('c', 0), ('a', 2)))
        db = ((('a', 3), ('b', 2), ('c', 0)))
        self.distributionTester(weightedChoice(da), db)

    def test_sample(self):
        """Test random sampling without replacement"""

        # returns experiment function with x = sample(c, draws)
        # experiment salt is a string version of c
        def sample(choices, draws, fast_sample=False):
            @experiment_decorator(','.join(map(str, choices)))
            def exp_func(e, i):
                if fast_sample:
                    e.x = FastSample(choices=choices, draws=draws, unit=i)
                else:
                    e.x = Sample(choices=choices, draws=draws, unit=i)
                self.assertTrue(len(e.x) == draws)
                return e
            return exp_func

        def listDistributionTester(func, value_mass, N=1000):
            value_density = TestRandomOperators.valueMassToDensity(value_mass)

            # compute N trials
            xs_list = [func(i=i).get('x') for i in six.moves.range(N)]

            # each xs is a row of the transpose of xs_list.
            # this is expected to have the same distribution as value_density
            for xs in zip(*xs_list):
                self.assertProbs(xs, value_density, float(N))

        listDistributionTester(
            sample([1, 2, 3], draws=3), ((1, 1), (2, 1), (3, 1)))
        listDistributionTester(
            sample([1, 2, 3], draws=2), ((1, 1), (2, 1), (3, 1)))
        listDistributionTester(
            sample([1, 2, 3], draws=2, fast_sample=True), ((1, 1), (2, 1), (3, 1)))
        listDistributionTester(
            sample(['a', 'a', 'b'], draws=3), (('a', 2), ('b', 1)))

        a = Assignment('assign_salt_a')
        a.old_sample = Sample(choices=[1, 2, 3, 4], draws=1, unit=1)
        new_sample = a.old_sample
        a.old_sample = FastSample(choices=[1, 2, 3, 4], draws=1, unit=1)
        self.assertTrue(len(a.old_sample), 1)
        self.assertTrue(len(new_sample), 1)
        self.assertTrue(a.old_sample != new_sample)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

import unittest

from planout.assignment import Assignment
from planout.ops.random import UniformChoice


class AssignmentTest(unittest.TestCase):
    tester_unit = 4
    tester_salt = 'test_salt'

    def test_set_get_constant(self):
        a = Assignment(self.tester_salt)
        a.foo = 12
        self.assertEqual(a.foo, 12)

    def test_set_get_uniform(self):
        a = Assignment(self.tester_salt)
        a.foo = UniformChoice(choices=['a', 'b'], unit=self.tester_unit)
        a.bar = UniformChoice(choices=['a', 'b'], unit=self.tester_unit)
        a.baz = UniformChoice(choices=['a', 'b'], unit=self.tester_unit)
        self.assertEqual(a.foo, 'b')
        self.assertEqual(a.bar, 'a')
        self.assertEqual(a.baz, 'a')

    def test_overrides(self):
        a = Assignment(self.tester_salt)
        a.set_overrides({'x': 42, 'y': 43})
        a.x = 5
        a.y = 6
        self.assertEqual(a.x, 42)
        self.assertEqual(a.y, 43)

    def test_custom_salt(self):
        a = Assignment(self.tester_salt)
        custom_salt = lambda x,y: '%s-%s' % (x,y)
        a.foo = UniformChoice(choices=list(range(8)), unit=self.tester_unit)
        self.assertEqual(a.foo, 7)

if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod, abstractproperty
from operator import itemgetter

from .experiment import Experiment, DefaultExperiment
from .ops.random import Sample, FastSample, RandomInteger
from .assignment import Assignment

# decorator for methods that assume assignments have been made


def requires_experiment(f):
    def wrapped_f(self, *args, **kwargs):
        if not self._experiment:
            self._assign_experiment()
        return f(self, *args, **kwargs)
    return wrapped_f


def requires_default_experiment(f):
    def wrapped_f(self, *args, **kwargs):
        if not self._default_experiment:
            self._assign_default_experiment()
        return f(self, *args, **kwargs)
    return wrapped_f


class Namespace(object):
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def add_experiment(self, name, exp_object, num_segments, **kwargs):
        pass

    @abstractmethod
    def remove_experiment(self, name):
        pass

    @abstractmethod
    def set_auto_exposure_logging(self, value):
        pass

    @abstractproperty
    def in_experiment(self):
        pass

    @abstractmethod
    def get(self, name, default):
        pass

    @abstractmethod
    def log_exposure(self, extras=None):
        pass

    @abstractmethod
    def log_event(self, event_type, extras=None):
        pass


class SimpleNamespace(Namespace):
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        self.name = self.__class__  # default name is the class name
        self.inputs = kwargs
        self.num_segments = None

        # dictionary mapping segments to experiment names
        self.segment_allocations = {}

        # dictionary mapping experiment names to experiment objects
        self.current_experiments = {}

        self._experiment = None          # memoized experiment object
        self._default_experiment = None  # memoized default experiment object
        self.default_experiment_class = DefaultExperiment
        self._in_experiment = False

        # setup name, primary key, number of segments, etc
        self.setup()
        self.available_segments = set(range(self.num_segments))

        # load namespace with experiments
        self.setup_experiments()

    @abstractmethod
    def setup(self):
        """Sets up experiment"""
        # Developers extending this class should set the following variables
        # self.name = 'sample namespace'
        # self.primary_unit = 'userid'
        # self.num_segments = 10000
        pass

    @abstractmethod
    def setup_experiments():
        # e.g.,
        # self.add_experiment('first experiment', Exp1, 100)
        pass

    @property
    def primary_unit(self):
        return self._primary_unit

    @primary_unit.setter
    def primary_unit(self, value):
        # later on we require that the primary key is a list, so we use
        # a setter to convert strings to a single element list
        if type(value) is list:
            self._primary_unit = value
        else:
            self._primary_unit = [value]

    def add_experiment(self, name, exp_object, segments):
        num_avail = len(self.available_segments)
        if num_avail < segments:
            print('error: %s segments requested, only %s available.' %
                  (segments, num_avail))
            return False
        if name in self.current_experiments:
            print('error: there is already an experiment called %s.' % name)
            return False

        # randomly select the given number of segments from all available
        # segments
        a = Assignment(self.name)
        if 'use_fast_sample' in self.inputs:
            a.sampled_segments = \
                FastSample(choices=list(self.available_segments),
                    draws=segments, unit=name)
        else:
            a.sampled_segments = \
                Sample(choices=list(self.available_segments),
                       draws=segments, unit=name)

        # assign each segment to the experiment name
        for segment in a.sampled_segments:
            self.segment_allocations[segment] = name
            self.available_segments.remove(segment)

        # associate the experiment name with an object
        self.current_experiments[name] = exp_object

    def remove_experiment(self, name):
        if name not in self.current_experiments:
            print('error: there is no experiment called %s.' % name)
            return False

        segments_to_free = \
            [s for s, n in self.segment_allocations.iteritems() if n == name]

        for segment in segments_to_free:
            del self.segment_allocations[segment]
            self.available_segments.add(segment)
        del self.current_experiments[name]

        return True

    def get_segment(self):
        # randomly assign primary unit to a segment
        a = Assignment(self.name)
        a.segment = RandomInteger(min=0, max=self.num_segments - 1,
                                  unit=itemgetter(*self.primary_unit)(self.inputs))
        return a.segment

    def _assign_experiment(self):
        "assign primary unit to an experiment"
        segment = self.get_segment()
        # is the unit allocated to an experiment?
        if segment in self.segment_allocations:
            experiment_name = self.segment_allocations[segment]
            experiment = self.current_experiments[
                experiment_name](**self.inputs)
            experiment.name = '%s-%s' % (self.name, experiment_name)
            experiment.salt = '%s.%s' % (self.name, experiment_name)
            self._experiment = experiment
            self._in_experiment = experiment.in_experiment
        # if the unit does not belong to an experiment, or the unit has been
        # disqualified from being in the assigned experiment, fall back on the
        # default experiment
        if not self._in_experiment:
            self._assign_default_experiment()

    def _assign_default_experiment(self):
        self._default_experiment = self.default_experiment_class(**self.inputs)

    @requires_default_experiment
    def default_get(self, name, default=None):
        return self._default_experiment.get(name, default)

    @property
    @requires_experiment
    def in_experiment(self):
        return self._in_experiment

    @in_experiment.setter
    def in_experiment(self, value):
        # in_experiment cannot be externally modified
        pass

    @requires_experiment
    def set_auto_exposure_logging(self, value):
        self._experiment.set_auto_exposure_logging(value)

    @requires_experiment
    def get(self, name, default=None):
        if self._experiment is None:
            return self.default_get(name, default)
        else:
            return self._experiment.get(name, self.default_get(name, default))

    @requires_experiment
    def log_exposure(self, extras=None):
        """Logs exposure to treatment"""
        if self._experiment is None:
            pass
        self._experiment.log_exposure(extras)

    @requires_experiment
    def log_event(self, event_type, extras=None):
        """Log an arbitrary event"""
        if self._experiment is None:
            pass
        self._experiment.log_event(event_type, extras)



# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

import json
import unittest

from planout.experiment import Experiment, SimpleInterpretedExperiment, ProductionExperiment
from planout.interpreter import Interpreter
from planout.ops.random import UniformChoice

global_log = []


class ExperimentTest(unittest.TestCase):

    def experiment_tester(self, exp_class, in_experiment=True):
        global global_log
        global_log = []

        e = exp_class(i=42)
        e.set_overrides({'bar': 42})
        params = e.get_params()

        self.assertTrue('foo' in params)
        self.assertEqual(params['foo'], 'b')

        # test to make sure overrides work correctly
        self.assertEqual(params['bar'], 42)

        # log should only have one entry, and should contain i as an input
        # and foo and bar as parameters
        if in_experiment:
            self.assertEqual(len(global_log), 1)
            self.validate_log(params, {
                'inputs': {'i': None},
                'params': {'foo': None, 'bar': None}
            })
        else:
            self.assertEqual(len(global_log), 0)

        # test to make sure experiment eligibility works correctly
        self.assertEqual(e.in_experiment, in_experiment)

    def validate_log(self, blob, expected_fields):
        # Expected field is a dictionary containing all of the expected keys
        # in the expected structure. Key values are ignored.
        blob = global_log[0]
        for field in expected_fields:
            self.assertTrue(field in blob)
            if expected_fields[field] is dict:
                self.assertTrue(self.validate_log(
                    blob[field],
                    expected_fields[field]
                ))
            else:
                self.assertTrue(field in blob)

    def test_vanilla_experiment(self):
        class TestVanillaExperiment(Experiment):

            def configure_logger(self):
                pass

            def log(self, stuff):
                global_log.append(stuff)

            def previously_logged(self):
                pass

            def setup(self):
                self.name = 'test_name'

            def assign(self, params, i):
                params.foo = UniformChoice(choices=['a', 'b'], unit=i)

        self.experiment_tester(TestVanillaExperiment)

    def test_vanilla_experiment_disabled(self):
        class TestVanillaExperiment(Experiment):

            def configure_logger(self):
                pass

            def log(self, stuff):
                global_log.append(stuff)

            def previously_logged(self):
                pass

            def setup(self):
                self.name = 'test_name'

            def assign(self, params, i):
                params.foo = UniformChoice(choices=['a', 'b'], unit=i)
                return 0

        self.experiment_tester(TestVanillaExperiment, False)

    # makes sure assignment only happens once
    def test_single_assignment(self):
        class TestSingleAssignment(Experiment):

            def configure_logger(self):
                pass

            def log(self, stuff):
                global_log.append(stuff)

            def previously_logged(self):
                pass

            def setup(self):
                self.name = 'test_name'

            def assign(self, params, i, counter):
                params.foo = UniformChoice(choices=['a', 'b'], unit=i)
                counter['count'] = counter.get('count', 0) + 1

        assignment_count = {'count': 0}
        e = TestSingleAssignment(i=10, counter=assignment_count)
        self.assertEqual(assignment_count['count'], 0)
        e.get('foo')
        self.assertEqual(assignment_count['count'], 1)
        e.get('foo')
        self.assertEqual(assignment_count['count'], 1)

    def test_interpreted_experiment(self):
        class TestInterpretedExperiment(SimpleInterpretedExperiment):

            def log(self, stuff):
                global_log.append(stuff)

            def loadScript(self):
                self.script = json.loads("""
                  {"op":"seq",
                   "seq": [
                    {"op":"set",
                     "var":"foo",
                     "value":{
                       "choices":["a","b"],
                       "op":"uniformChoice",
                       "unit": {"op": "get", "var": "i"}
                       }
                    },
                    {"op":"set",
                     "var":"bar",
                     "value": 41
                    }
                   ]}
                """)

        self.experiment_tester(TestInterpretedExperiment)

    def test_disabled_interpreted_experiment(self):
        class TestInterpretedDisabled(SimpleInterpretedExperiment):
            def log(self, stuff):
                global_log.append(stuff)

            def loadScript(self):
                self.script = json.loads("""
                    {
                     "op": "seq",
                     "seq": [
                      {
                       "op": "set",
                       "var": "foo",
                       "value": "b"
                      },
                      {
                       "op": "return",
                       "value": false
                      }
                     ]
                    }
                    """)

        self.experiment_tester(TestInterpretedDisabled, False)

    def test_short_circuit_exposure_logging(self):
        class TestNoExposure(Experiment):

            def configure_logger(self):
                pass

            def log(self, stuff):
                global_log.append(stuff)

            def previously_logged(self):
                pass

            def setup(self):
                self.name = 'test_name'

            def assign(self, params, i):
                params.foo = UniformChoice(choices=['a', 'b'], unit=i)
                return False
        self.experiment_tester(TestNoExposure, False)

        class TestNoExposure(Experiment):

            def configure_logger(self):
                pass

            def log(self, stuff):
                global_log.append(stuff)

            def previously_logged(self):
                pass

            def setup(self):
                self.name = 'test_name'

            def assign(self, params, i):
                params.foo = UniformChoice(choices=['a', 'b'], unit=i)
                return True
        self.experiment_tester(TestNoExposure, True)

    def test_demo_production_experiment(self):
        class TestDemoProductionExperiment(ProductionExperiment):
            def configure_logger(self):
                pass

            def log(self, stuff):
                global_log.append(stuff)

            def previously_logged(self):
                pass

            def setup(self):
                self.name = 'test_name'

            def assign(self, params, i):
                params.foo = UniformChoice(choices=['a', 'b'], unit=i)

            def get_param_names(self):
                return ['foo']
        
        e = TestDemoProductionExperiment(i=42)
        self.assertEqual(e.get('nobar'), None)
        self.assertEqual(len(global_log), 0)
        self.assertEqual(e.get('foo'), 'b')
        self.assertEqual(len(global_log), 1)

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

import logging
import time
import re
import json
import inspect
import hashlib
from abc import ABCMeta, abstractmethod
import six
import __main__ as main

from .assignment import Assignment
from .interpreter import Interpreter


# decorator for methods that assume assignments have been made
def requires_assignment(f):
    def wrapped_f(self, *args, **kwargs):
        if not self._assigned:
            self._assign()
        return f(self, *args, **kwargs)
    return wrapped_f

# decorator for methods that should be exposure logged


def requires_exposure_logging(f):
    def wrapped_f(self, *args, **kwargs):
        if self._auto_exposure_log and not self._exposure_logged:
            should_log_exposure = True
            if f.__name__ == 'get' and hasattr(self, 'get_param_names') and callable(self.get_param_names):
                params = self.get_param_names()
                should_log_exposure = any(i in params for i in args)
            if should_log_exposure:
                self.log_exposure()
        return f(self, *args, **kwargs)
    return wrapped_f


class Experiment(object):

    """Abstract base class for PlanOut experiments"""
    __metaclass__ = ABCMeta

    logger_configured = False

    def __init__(self, **inputs):
        self.inputs = inputs           # input data
        # True when assignments have been exposure logged
        self._exposure_logged = False
        self._salt = None              # Experiment-level salt

        # Determines whether or not exposure should be logged
        self._in_experiment = True

        # use the name of the class as the default name
        self._name = self.__class__.__name__

        # auto-exposure logging is enabled by default
        self._auto_exposure_log = True

        self.setup()                   # sets name, salt, etc.

        self._assignment = Assignment(self.salt)
        self._assigned = False

    def _assign(self):
        """Assignment and setup that only happens when we need to log data"""
        self.configure_logger()  # sets up loggers

        #consumers can optionally return False from assign if they don't want exposure to be logged
        assign_val = self.assign(self._assignment, **self.inputs)
        self._in_experiment = True if assign_val or assign_val is None else False
        self._checksum = self.checksum()
        self._assigned = True

    def setup(self):
        """Set experiment attributes, e.g., experiment name and salt."""
        # If the experiment name is not specified, just use the class name
        pass

    def set_overrides(self, value):
        """Sets variables that are to remain fixed during execution."""
        # note that setting this will overwrite inputs to the experiment
        self._assignment.set_overrides(value)
        o = self._assignment.get_overrides()
        for var in o:
            if var in self.inputs:
                self.inputs[var] = o[var]

    @property
    def in_experiment(self):
        return self._in_experiment

    @property
    def salt(self):
        # use the experiment name as the salt if the salt is not set
        return self._salt if self._salt else self.name

    @salt.setter
    def salt(self, value):
        self._salt = value
        if hasattr(self, '_assignment'):
            self._assignment.experiment_salt = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = re.sub(r'\s+', '-', value)
        if hasattr(self, '_assignment'):
            self._assignment.experiment_salt = self.salt

    @abstractmethod
    def assign(params, **kwargs):
        """Returns evaluated PlanOut mapper with experiment assignment"""
        pass

    @requires_assignment
    def __asBlob(self, extras={}):
        """Dictionary representation of experiment data"""
        d = {
            'name': self.name,
            'time': int(time.time()),
            'salt': self.salt,
            'inputs': self.inputs,
            'params': dict(self._assignment),
        }
        for k in extras:
            d[k] = extras[k]
        if self._checksum:
            d['checksum'] = self._checksum
        return d

    def checksum(self):
        # if we're running from a file and want to detect if the experiment
        # file has changed
        if hasattr(main, '__file__'):
            # src doesn't count first line of code, which includes function
            # name
            src = ''.join(inspect.getsourcelines(self.assign)[0][1:])
            if not isinstance(src, six.binary_type):
                src = src.encode("ascii")
            return hashlib.sha1(src).hexdigest()[:8]
        # if we're running in an interpreter, don't worry about it
        else:
            return None

    # we should probably get rid of this public interface
    @property
    def exposure_logged(self):
        return self._exposure_logged

    def set_auto_exposure_logging(self, value):
        """
        Disables / enables auto exposure logging (enabled by default).
        """
        self._auto_exposure_log = value

    @requires_assignment
    @requires_exposure_logging
    def get_params(self):
        """
        Get all PlanOut parameters. Triggers exposure log.
        """
        # In general, this should only be used by custom loggers.
        return dict(self._assignment)

    @requires_assignment
    @requires_exposure_logging
    def get(self, name, default=None):
        """
        Get PlanOut parameter (returns default if undefined). Triggers exposure log.
        """
        return self._assignment.get(name, default)

    @requires_assignment
    @requires_exposure_logging
    def __str__(self):
        """
        String representation of exposure log data. Triggers exposure log.
        """
        return str(self.__asBlob())

    def log_exposure(self, extras=None):
        """Logs exposure to treatment"""
        if not self._in_experiment:
            return
        self._exposure_logged = True
        self.log_event('exposure', extras)

    def log_event(self, event_type, extras=None):
        """Log an arbitrary event"""
        if not self._in_experiment:
            return
        if extras:
            extra_payload = {'event': event_type, 'extra_data': extras.copy()}
        else:
            extra_payload = {'event': event_type}
        self.log(self.__asBlob(extra_payload))

    @abstractmethod
    def configure_logger(self):
        """Set up files, database connections, sockets, etc for logging."""
        pass

    @abstractmethod
    def log(self, data):
        """Log experimental data"""
        pass

    @abstractmethod
    def previously_logged(self):
        """Check if the input has already been logged.
           Gets called once during in the constructor."""
        # For high-use applications, one might have this method to check if
        # there is a memcache key associated with the checksum of the
        # inputs+params
        pass


class DefaultExperiment(Experiment):

    """
    Dummy experiment which has no logging. Default experiments used by namespaces
    should inherent from this class.
    """

    def configure_logger(self):
        pass  # we don't log anything when there is no experiment

    def log(self, data):
        pass

    def previously_logged(self):
        return True

    def assign(self, params, **kwargs):
        # more complex default experiments can override this method
        params.update(self.get_default_params())

    def get_default_params(self):
        """
        Default experiments that are just key-value stores should
        override this method."""
        return {}


class SimpleExperiment(Experiment):

    """Simple experiment base class which exposure logs to a file"""

    __metaclass__ = ABCMeta
    # We only want to set up the logger once, the first time the object is
    # instantiated. We do this by maintaining this class variable.
    logger = {}
    log_file = {}

    def configure_logger(self):
        """Sets up logger to log to a file"""
        my_logger = self.__class__.logger
        # only want to set logging handler once for each experiment (name)
        if self.name not in self.__class__.logger:
            if self.name not in self.__class__.log_file:
                self.__class__.log_file[self.name] = '%s.log' % self.name
            file_name = self.__class__.log_file[self.name]
            my_logger[self.name] = logging.getLogger(self.name)
            my_logger[self.name].setLevel(logging.INFO)
            my_logger[self.name].addHandler(logging.FileHandler(file_name))
            my_logger[self.name].propagate = False

    def log(self, data):
        """Logs data to a file"""
        self.__class__.logger[self.name].info(json.dumps(data))

    def set_log_file(self, path):
        self.__class__.log_file[self.name] = path

    def previously_logged(self):
        """Check if the input has already been logged.
           Gets called once during in the constructor."""
        # SimpleExperiment doesn't connect with any services, so we just assume
        # that if the object is a new instance, this is the first time we are
        # seeing the inputs/outputs given.
        return False


class SimpleInterpretedExperiment(SimpleExperiment):

    """A variant of SimpleExperiment that loads data from a given script"""
    __metaclass__ = ABCMeta

    def loadScript(self):
        """loads deserialized PlanOut script to be executed by the interpreter"""
        # This method should set self.script to a dictionary-based representation
        # of a PlanOut script. Most commonly, this method would retreive a
        # JSON-encoded string from a database or file, e.g.
        # self.script = json.loads(open("myfile").read())
        # If constructing experiments on the fly, one can alternatively set the
        # self.script instance variable
        pass

    def assign(self, params, **kwargs):
        self.loadScript()  # lazily load script
        # self.script must be a dictionary
        assert hasattr(self, 'script') and type(self.script) == dict

        interpreterInstance = Interpreter(
            self.script,
            self.salt,
            kwargs,
            params
        )
        # execute script
        results = interpreterInstance.get_params()
        # insert results into param object dictionary
        params.update(results)
        return interpreterInstance.in_experiment

    def checksum(self):
        # self.script must be a dictionary
        assert hasattr(self, 'script') and type(self.script) == dict
        src = json.dumps(self.script)

        if not isinstance(src, six.binary_type):
            src = src.encode("ascii")
        return hashlib.sha1(src).hexdigest()[:8]

class ProductionExperiment(Experiment):

    """ 
    A variant of SimpleExperiment that verifies that exposure is only logged 
    when a valid parameter is fetched via the get method
    """
    
    __metaclass__ = ABCMeta


    """Returns a list of assignment parameter values that this experiment can take"""
    @abstractmethod
    def get_param_names(self):
        pass




# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

import unittest
import six

from planout.interpreter import (
    Interpreter,
)


class TestBasicOperators(unittest.TestCase):

    def runConfig(self, config, init={}):
        e = None
        e = Interpreter(config, 'test_salt', init)
        return e.get_params()

    def run_config_single(self, config):
        x_config = {'op': 'set', 'var': 'x', 'value': config}
        return self.runConfig(x_config)['x']

    def test_set(self):
        """Test setter"""
        # returns experiment object with probability p
        c = {'op': 'set', 'value': 'x_val', 'var': 'x'}
        d = self.runConfig(c)
        self.assertEquals({'x': 'x_val'}, d)

    def test_seq(self):
        """Test sequence"""
        config = {'op': 'seq', 'seq': [
            {'op': 'set', 'value': 'x_val', 'var': 'x'},
            {'op': 'set', 'value': 'y_val', 'var': 'y'}
        ]}
        d = self.runConfig(config)
        self.assertEquals({'x': 'x_val', 'y': 'y_val'}, d)

    def test_array(self):
        arr = [4, 5, 'a']
        a = self.run_config_single({'op': 'array', 'values': arr})
        self.assertEquals(arr, a)

    def test_map(self):
        my_map = {'a': 2, 'b': 'c', 'd': False}
        m = self.run_config_single({'op': 'map', 'a': 2, 'b': 'c', 'd': False})
        self.assertEquals(my_map, m)
        
        my_map = {}
        m = self.run_config_single({'op': 'map'})
        self.assertEquals(my_map, m)

    def test_cond(self):
        getInput = lambda i, r: {'op': 'equals', 'left': i, 'right': r}
        testIf = lambda i: self.runConfig({
            'op': 'cond',
            'cond': [
                {'if': getInput(i, 0),
                 'then': {'op': 'set', 'var': 'x', 'value': 'x_0'}},
                {'if': getInput(i, 1),
                 'then': {'op': 'set', 'var': 'x', 'value': 'x_1'}}
            ]
        })
        self.assertEquals({'x': 'x_0'}, testIf(0))
        self.assertEquals({'x': 'x_1'}, testIf(1))

    def test_get(self):
        d = self.runConfig({
            'op': 'seq',
            'seq': [
                {'op': 'set', 'var': 'x', 'value': 'x_val'},
                {'op': 'set', 'var': 'y', 'value': {'op': 'get', 'var': 'x'}}
            ]
        })
        self.assertEquals({'x': 'x_val', 'y': 'x_val'}, d)

    def test_index(self):
        array_literal = [10, 20, 30]
        dict_literal = {'a': 42, 'b': 43}

        # basic indexing works with array literals
        x = self.run_config_single(
            {'op': 'index', 'index': 0, 'base': array_literal}
        )
        self.assertEquals(x, 10)

        x = self.run_config_single(
            {'op': 'index', 'index': 2, 'base': array_literal}
        )
        self.assertEquals(x, 30)

        # basic indexing works with dictionary literals
        x = self.run_config_single(
            {'op': 'index', 'index': 'a', 'base': dict_literal}
        )
        self.assertEquals(x, 42)

        # invalid indexes are mapped to None
        x = self.run_config_single(
            {'op': 'index', 'index': 6, 'base': array_literal}
        )
        self.assertEquals(x, None)

        # invalid indexes are mapped to None
        x = self.run_config_single(
            {'op': 'index', 'index': 'c', 'base': dict_literal}
        )
        self.assertEquals(x, None)

        # non literals also work
        x = self.run_config_single({
            'op': 'index',
            'index': 2,
            'base': {'op': 'array', 'values': array_literal}
        })
        self.assertEquals(x, 30)

    def test_coalesce(self):
        x = self.run_config_single({'op': 'coalesce', 'values': [None]})
        self.assertEquals(x, None)

        x = self.run_config_single(
            {'op': 'coalesce', 'values': [None, 42, None]})
        self.assertEquals(x, 42)

        x = self.run_config_single(
            {'op': 'coalesce', 'values': [None, None, 43]})
        self.assertEquals(x, 43)

    def test_length(self):
        arr = list(range(5))
        length_test = self.run_config_single({'op': 'length', 'value': arr})
        self.assertEquals(len(arr), length_test)
        length_test = self.run_config_single({'op': 'length', 'value': []})
        self.assertEquals(0, length_test)
        length_test = self.run_config_single({'op': 'length', 'value':
                                              {'op': 'array', 'values': arr}
                                              })
        self.assertEquals(length_test, len(arr))

    def test_not(self):
        # test not
        x = self.run_config_single({'op': 'not', 'value': 0})
        self.assertEquals(True, x)
        x = self.run_config_single({'op': 'not', 'value': False})
        self.assertEquals(True, x)

        x = self.run_config_single({'op': 'not', 'value': 1})
        self.assertEquals(False, x)
        x = self.run_config_single({'op': 'not', 'value': True})
        self.assertEquals(False, x)

    def test_or(self):
        x = self.run_config_single({
            'op': 'or',
            'values': [0, 0, 0]})
        self.assertEquals(False, x)

        x = self.run_config_single({
            'op': 'or',
            'values': [0, 0, 1]})
        self.assertEquals(True, x)

        x = self.run_config_single({
            'op': 'or',
            'values': [False, True, False]})
        self.assertEquals(True, x)

    def test_and(self):
        x = self.run_config_single({
            'op': 'and',
            'values': [1, 1, 0]})
        self.assertEquals(False, x)

        x = self.run_config_single({
            'op': 'and',
            'values': [0, 0, 1]})
        self.assertEquals(False, x)

        x = self.run_config_single({
            'op': 'and',
            'values': [True, True, True]})
        self.assertEquals(True, x)

    def test_commutative(self):
        # test commutative arithmetic operators
        arr = [33, 7, 18, 21, -3]

        min_test = self.run_config_single({'op': 'min', 'values': arr})
        self.assertEquals(min(arr), min_test)

        max_test = self.run_config_single({'op': 'max', 'values': arr})
        self.assertEquals(max(arr), max_test)

        sum_test = self.run_config_single({'op': 'sum', 'values': arr})
        self.assertEquals(sum(arr), sum_test)

        product_test = self.run_config_single({'op': 'product', 'values': arr})
        self.assertEquals(six.moves.reduce(lambda x, y: x * y, arr), product_test)

    def test_binary_ops(self):
        eq = self.run_config_single({'op': 'equals', 'left': 1, 'right': 2})
        self.assertEquals(1 == 2, eq)
        eq = self.run_config_single({'op': 'equals', 'left': 2, 'right': 2})
        self.assertEquals(2 == 2, eq)
        gt = self.run_config_single({'op': '>', 'left': 1, 'right': 2})
        self.assertEquals(1 > 2, gt)
        lt = self.run_config_single({'op': '<', 'left': 1, 'right': 2})
        self.assertEquals(1 < 2, lt)
        gte = self.run_config_single({'op': '>=', 'left': 2, 'right': 2})
        self.assertEquals(2 >= 2, gte)
        gte = self.run_config_single({'op': '>=', 'left': 1, 'right': 2})
        self.assertEquals(1 >= 2, gte)
        lte = self.run_config_single({'op': '<=', 'left': 2, 'right': 2})
        self.assertEquals(2 <= 2, lte)
        mod = self.run_config_single({'op': '%', 'left': 11, 'right': 3})
        self.assertEquals(11 % 3, mod)
        div = self.run_config_single({'op': '/', 'left': 3, 'right': 4})
        self.assertEquals(0.75, div)

    def test_return(self):
        def return_runner(return_value):
            config = {
                "op": "seq",
                "seq": [
                    {
                      "op": "set",
                      "var": "x",
                      "value": 2
                    },
                    {
                        "op": "return",
                        "value": return_value
                    },
                    {
                        "op": "set",
                        "var": "y",
                        "value": 4
                    }
                ]
            }
            e = Interpreter(config, 'test_salt')
            return e

        i = return_runner(True)
        self.assertEquals({'x': 2}, i.get_params())
        self.assertEquals(True, i.in_experiment)
        i = return_runner(42)
        self.assertEquals({'x': 2}, i.get_params())
        self.assertEquals(True, i.in_experiment)
        i = return_runner(False)
        self.assertEquals({'x': 2}, i.get_params())
        self.assertEquals(False, i.in_experiment)
        i = return_runner(0)
        self.assertEquals({'x': 2}, i.get_params())
        self.assertEquals(False, i.in_experiment)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

import json
import unittest

from planout.interpreter import Interpreter


class InterpreterTest(unittest.TestCase):
    compiled = json.loads("""
{"op":"seq","seq":[{"op":"set","var":"group_size","value":{"choices":{"op":"array","values":[1,10]},"unit":{"op":"get","var":"userid"},"op":"uniformChoice"}},{"op":"set","var":"specific_goal","value":{"p":0.8,"unit":{"op":"get","var":"userid"},"op":"bernoulliTrial"}},{"op":"cond","cond":[{"if":{"op":"get","var":"specific_goal"},"then":{"op":"seq","seq":[{"op":"set","var":"ratings_per_user_goal","value":{"choices":{"op":"array","values":[8,16,32,64]},"unit":{"op":"get","var":"userid"},"op":"uniformChoice"}},{"op":"set","var":"ratings_goal","value":{"op":"product","values":[{"op":"get","var":"group_size"},{"op":"get","var":"ratings_per_user_goal"}]}}]}}]}]}
  """)
    interpreter_salt = 'foo'

    def test_interpreter(self):
        proc = Interpreter(
            self.compiled, self.interpreter_salt, {'userid': 123454})
        params = proc.get_params()
        self.assertEqual(proc.get_params().get('specific_goal'), 1)
        self.assertEqual(proc.get_params().get('ratings_goal'), 320)

    def test_interpreter_overrides(self):
        # test overriding a parameter that gets set by the experiment
        proc = Interpreter(
            self.compiled, self.interpreter_salt, {'userid': 123454})
        proc.set_overrides({'specific_goal': 0})
        self.assertEqual(proc.get_params().get('specific_goal'), 0)
        self.assertEqual(proc.get_params().get('ratings_goal'), None)

        # test to make sure input data can also be overridden
        proc = Interpreter(
            self.compiled, self.interpreter_salt, {'userid': 123453})
        proc.set_overrides({'userid': 123454})
        self.assertEqual(proc.get_params().get('specific_goal'), 1)

    def test_register_ops(self):
        from planout.ops.base import PlanOutOpCommutative
        class CustomOp(PlanOutOpCommutative):
            def commutativeExecute(self, values):
                return sum(values)

        custom_op_script = {"op":"seq","seq":[{"op":"set","var":"x","value":{"values":[2,4],"op":"customOp"}}]}
        proc = Interpreter(
            custom_op_script, self.interpreter_salt, {'userid': 123454})

        proc.register_operators({'customOp': CustomOp})
        self.assertEqual(proc.get_params().get('x'), 6)



if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2014, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

import logging
from abc import ABCMeta, abstractmethod
import six

from .utils import Operators


class PlanOutOp(object):

    """Abstract base class for PlanOut Operators"""
    __metaclass__ = ABCMeta
    # all PlanOut operator have some set of args that act as required and
    # optional arguments

    def __init__(self, **args):
        self.args = args

    # all PlanOut operators must implement execute
    @abstractmethod
    def execute(self, mapper):
        pass

    def prettyArgs(self):
        return Operators.prettyParamFormat(self.args)

    def pretty(self):
        return '%s(%s)' % (self.args['op'], self.prettyArgs())

    def getArgMixed(self, name):
        assert name in self.args, \
            "%s: missing argument: %s." % (self.__class__, name)
        return self.args[name]

    def getArgInt(self, name):
        arg = self.getArgMixed(name)
        assert isinstance(arg, six.integer_types), \
            "%s: %s must be an int." % (self.__class__, name)
        return arg

    def getArgFloat(self, name):
        arg = self.getArgMixed(name)
        assert isinstance(arg, (six.integer_types, float)), \
            "%s: %s must be a number." % (self.__class__, name)
        return float(arg)

    def getArgString(self, name):
        arg = self.getArgMixed(name)
        assert isinstance(arg, six.string_types), \
            "%s: %s must be a string." % (self.__class__, name)
        return arg

    def getArgNumeric(self, name):
        arg = self.getArgMixed(name)
        assert isinstance(arg, (six.integer_types, float)), \
            "%s: %s must be a numeric." % (self.__class__, name)
        return arg

    def getArgList(self, name):
        arg = self.getArgMixed(name)
        assert isinstance(arg, (list, tuple)), \
            "%s: %s must be a list." % (self.__class__, name)
        return arg

    def getArgMap(self, name):
        arg = self.getArgMixed(name)
        assert isinstance(arg, dict), \
            "%s: %s must be a map." % (self.__class__, name)
        return arg

    def getArgIndexish(self, name):
        arg = self.getArgMixed(name)
        assert isinstance(arg, (dict, list, tuple)), \
            "%s: %s must be a map or list." % (self.__class__, name)
        return arg

# PlanOutOpSimple is the easiest way to implement simple operators.
# The class automatically evaluates the values of all args passed in via
# execute(), and stores the PlanOut mapper object and evaluated
# args as instance variables.  The user can then extend PlanOutOpSimple
# and implement simpleExecute().

class PlanOutOpSimple(PlanOutOp):
    __metaclass__ = ABCMeta

    def execute(self, mapper):
        self.mapper = mapper
        parameter_names = self.args.keys()
        for param in parameter_names:
            self.args[param] = mapper.evaluate(self.args[param])
        return self.simpleExecute()


class PlanOutOpBinary(PlanOutOpSimple):
    __metaclass__ = ABCMeta

    def simpleExecute(self):
        return self.binaryExecute(
            self.getArgMixed('left'),
            self.getArgMixed('right'))

    def pretty(self):
        return '%s %s %s' % (
            Operators.pretty(self.args['left']),
            self.getInfixString(),
            Operators.pretty(self.args['right']))

    def getInfixString(self):
        return self.args['op']

    @abstractmethod
    def binaryExecute(self, left, right):
        pass


class PlanOutOpUnary(PlanOutOpSimple):
    __metaclass__ = ABCMeta

    def simpleExecute(self):
        return self.unaryExecute(self.getArgMixed('value'))

    def pretty(self):
        return self.getUnaryString + Operators.pretty(self.getArgMixed('value'))

    def getUnaryString(self):
        return self.args['op']

    @abstractmethod
    def unaryExecute(self, value):
        pass


class PlanOutOpCommutative(PlanOutOpSimple):
    __metaclass__ = ABCMeta

    def simpleExecute(self):
        assert ('values' in self.args), "expected argument 'values'"
        return self.commutativeExecute(self.getArgList('values'))

    def pretty(self):
        values = Operators.strip_array(self.getArgList('values'))
        if type(values) is list:
            pretty_values = ', '.join([Operators.pretty(i) for i in values])
        else:
            pretty_values = Operators.pretty(values)

        return '%s(%s)' % (self.getCommutativeString(), pretty_values)

    def getCommutativeString(self):
        return self.args['op']

    @abstractmethod
    def commutativeExecute(self, values):
        pass
